from amaranth import Module
from amaranth.lib import wiring, data
from amaranth.lib.memory import Memory
from amaranth.sim import Simulator
import pytest

from groovylight.common import Rgb888Layout, Rgb666Layout

from groovylight.hub75 import (
    DisplayClock,
    Hub75Coordinator,
    Hub75DataDriver,
    SwapBuffer,
)


def test_swapbuffer():
    dut = SwapBuffer(Rgb666Layout, 512)
    sim = Simulator(dut)
    sim.add_clock(1e-6)

    async def testbench(ctx):
        init_color = {"red": 0, "green": 0, "blue": 0}
        test_color = {"red": 8, "green": 8, "blue": 8}
        ctx.set(dut.selector, 0)
        ctx.set(dut.write_port.addr, 1)
        ctx.set(dut.read_port.addr, 1)
        ctx.set(dut.write_port.data, test_color)
        await ctx.tick()
        # assert that the read port addr 1 = 0
        assert ctx.get(dut.read_port.data) == init_color
        # swap buffer
        ctx.set(dut.selector, 1)
        await ctx.tick().repeat(
            2
        )  # takes two clocks after switching selector to output data.
        assert ctx.get(dut.read_port.data) == test_color

    # TODO: add more assertions/verification
    sim.add_testbench(testbench)
    with sim.write_vcd("output.vcd"):
        sim.run_until(1e-6 * 1000)


def test_datadriver():
    # the string driver test must
    # 1. finish
    # 2. strobe through all of the data in the array
    # 3. slice the correct bit from the data.
    memdata = [{"red": x, "green": x, "blue": x} for x in range(256)]
    m = Module()
    m.submodules.dut = dut = Hub75DataDriver()
    m.submodules.mem = mem = Memory(shape=Rgb888Layout, depth=256, init=memdata)
    m.submodules.clocker = clocker = DisplayClock()
    m.d.comb += [
        clocker.start.eq(dut.start),
        clocker.done.eq(dut.done),
    ]

    port = mem.read_port()

    wiring.connect(m, port, dut.bram)

    async def testbench(ctx):
        # select a bit, strobe start, read values, test against known.
        ctx.set(dut.bcm_select, 7)
        ctx.set(dut.start, 1)
        await ctx.tick()
        ctx.set(dut.start, 0)
        assert ctx.get(dut.bram.en) == 1
        await ctx.tick().until(dut.done == 1)

    async def rgbtest(ctx):
        await ctx.tick().until(dut.start == 1)
        counter = 127
        bitslice = 7
        async for _, rgb0, rgb1 in ctx.posedge(clocker.clk).sample(dut.data.rgb0, dut.data.rgb1):
            assert counter >= 0, "should not do more than 128 clocks"
            e0 = ctx.get(mem.data[counter << 1])
            e1 = ctx.get(mem.data[(counter << 1) + 1])
            print(counter)
            for r, e in [(rgb0, e0), (rgb1, e1)]:
                assert r.red == (e.red >> bitslice) & 1
                assert r.green == (e.green >> bitslice) & 1
                assert r.blue == (e.blue >> bitslice) & 1
            counter = counter - 1


    sim = Simulator(m)
    sim.add_clock(1e-6)
    sim.add_testbench(testbench)
    sim.add_testbench(rgbtest, background=True)

    with sim.write_vcd("output.vcd"):
        sim.run()
        # sim.run_until(1e-6 * 4000)


@pytest.mark.skip()
def test_hub75_coordinator():
    m = Module()
    m.submodules.dut = dut = Hub75Coordinator(1)

    sim = Simulator(m)
    sim.add_clock(1e-6)

    with sim.write_vcd("output.vcd"):
        sim.run_until(1e-6 * 1000)
